/*! autodux v3.0.0 by Eric Elliott */
var get = require('lodash/fp/get');
var capitalize = require('lodash/upperFirst');

var id = function (x) {
  return x;
};
var selectIf = function (predicate) {
  return function (x) {
    return predicate(x) && x;
  };
};
var isFunction = function (f) {
  return typeof f === 'function';
};
var selectFunction = selectIf(isFunction);

// # Selector creation:
var toGetter = function (s) {
  return 'get' + capitalize(s);
};
var sliceSelector = function (slice, fn) {
  return function (state) {
    return fn(state[slice], state);
  };
};

var createInitSelectors = function (slice, initial) {
  return Object.keys(initial).reduce(function (obj, selector) {
    var _Object$assign, _Object$assign2;

    return slice ? Object.assign(obj, (_Object$assign = {}, _Object$assign[toGetter(selector)] = sliceSelector(slice, function (state) {
      return get(selector, state);
    }), _Object$assign)) : Object.assign(obj, (_Object$assign2 = {}, _Object$assign2[toGetter(selector)] = function (state) {
      return get(selector, state);
    }, _Object$assign2));
  }, {});
};

var createSliceSelectors = function (slice, selectors) {
  var _Object$assign5;

  return Object.assign(Object.keys(selectors).reduce(function (obj, selector) {
    var _Object$assign3, _Object$assign4;

    return slice ? Object.assign(obj, (_Object$assign3 = {}, _Object$assign3[selector] = sliceSelector(slice, selectors[selector]), _Object$assign3)) : Object.assign(obj, (_Object$assign4 = {}, _Object$assign4[selector] = selectors[selector], _Object$assign4));
  }, {}), (_Object$assign5 = {}, _Object$assign5[toGetter(slice)] = sliceSelector(slice, id), _Object$assign5));
};
// /selector creation

// # Action creation
var getActionName = function (key) {
  return 'set' + capitalize(key);
};

var createAction = function (slice, action, key) {
  var type = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : slice + '/' + action;
  return {
    create: Object.assign(function (payload) {
      return {
        type: type,
        payload: payload
      };
    }, { type: type }),

    reducer: function reducer(state, _ref) {
      var _ref2;

      var payload = _ref.payload;
      return Object.assign({}, state, key ? (_ref2 = {}, _ref2[key] = payload, _ref2) : payload);
    }
  };
};

var getInitialActions = function (slice, initial) {
  return Object.keys(initial).reduce(function (o, key) {
    var _Object$assign6;

    var action = getActionName(key);
    return Object.assign(o, (_Object$assign6 = {}, _Object$assign6[action] = createAction(slice, action, key), _Object$assign6));
  }, {});
};

var addDefaultActions = function (slice, initial, actions) {
  var _Object$assign7;

  var initialActions = getInitialActions(slice, initial);
  var action = getActionName(slice);

  return Object.assign({}, (_Object$assign7 = {}, _Object$assign7[action] = createAction(slice, action), _Object$assign7), initialActions, actions);
};

var createMappedActions = function (slice, actions) {
  return Object.keys(actions).reduce(function (obj, action) {
    var _Object$assign8;

    return Object.assign({}, obj, (_Object$assign8 = {}, _Object$assign8[action] = Object.assign(function () {
      var _actions$action;

      return {
        type: slice + '/' + action,
        payload: typeof actions[action].create === 'function' ? (_actions$action = actions[action]).create.apply(_actions$action, arguments) : arguments.length <= 0 ? undefined : arguments[0]
      };
    }, { type: slice + '/' + action }), _Object$assign8));
  }, {});
};
// /action creation

var autodux = function () {
  var _ref3 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      _ref3$initial = _ref3.initial,
      initial = _ref3$initial === undefined ? '' : _ref3$initial,
      _ref3$actions = _ref3.actions,
      actions = _ref3$actions === undefined ? {} : _ref3$actions,
      _ref3$selectors = _ref3.selectors,
      selectors = _ref3$selectors === undefined ? {} : _ref3$selectors,
      _ref3$slice = _ref3.slice,
      slice = _ref3$slice === undefined ? '' : _ref3$slice;

  var allSelectors = Object.assign({}, createInitSelectors(slice, initial), createSliceSelectors(slice, selectors));

  var allActions = createMappedActions(slice, addDefaultActions(slice, initial, actions));

  return {
    initial: initial, reducer: function reducer() {
      var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : initial;

      var _ref4 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
          type = _ref4.type,
          payload = _ref4.payload;

      var _ref5 = type ? type.split('/') : 'unknown/unknown'.split('/'),
          namespace = _ref5[0],
          subType = _ref5[1];

      var defaultActions = addDefaultActions(slice, initial, {});

      // Look for reducer with top-to-bottom precedence.
      // Fall back to default actions, then undefined.
      // The actions[subType] key can be a function
      // or an object, so we only select the value
      // if it's a function:
      var actionReducer = [get(subType + '.reducer', actions), actions[subType], get(subType + '.reducer', defaultActions)].reduceRight(function (f, v) {
        return selectFunction(v) || f;
      });

      return namespace === slice && (actions[subType] || defaultActions[subType]) ? actionReducer ? actionReducer(state, payload) : Object.assign({}, state, payload) : state;
    }, slice: slice,
    selectors: allSelectors, actions: allActions
  };
};

module.exports = autodux;
module.exports.id = id;

module.exports.assign = function (key) {
  return function (state, payload) {
    var _Object$assign9;

    return Object.assign({}, state, (_Object$assign9 = {}, _Object$assign9[key] = payload, _Object$assign9));
  };
};
//# sourceMappingURL=index-esm.js.map
